#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#define __USE_GNU
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <ctype.h>
#include <assert.h>
#include "fishguts.h"
#include "fish.h"

#define MAX_NEIGHBORS 12
#define MAXPACKET 1024
#define MINIMUM_TIME_BETWEEN_TWO_SENDS .01

#define MIN(a,b) (((a)>(b)) ? (b) : (a))
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define MAX3(a,b,c) (((a)>(b)) ? (MAX(a,c)) : (MAX(a,b)))
#define InRange(low, val, up) (((low) <= (val)) && ((val) <= (up)))
#define PRINT_AND_BAIL(x) do { printpacket(pkt, packet_length); \
  errorabort(x); } while(0);

/*@-booltype boolean@*/
typedef int boolean;
#define FALSE (0)
#define TRUE (!(FALSE))

/* yeah; we store the node's address, but we don't have to. */
static int self_address = -1;

static boolean init = FALSE;
/* the sockets are -1 so that join can test whether they're
   valid to protect against being called twice */
static int sockfishhead = -1;
static int sockfishbits = -1;
static struct sockaddr_in headaddr;
static struct sockaddr_in bitsaddr;
static time_t last_send_time;  

/* +1 to allow null termination */
static struct neighbor neighbors[MAX_NEIGHBORS+1];

static receive_handler recvhandler;
static keyboard_handler keybhandler;

void read_neighbor_list(void); 
int find_by_sock_address(const struct sockaddr_in addr);
void packet_verification(const char *pkt, int len);
void printpacket(const struct packet *pkt, int len);

static int fish_constructmask(fd_set *prd, int max);
static void fish_processbindings(fd_set *prd);

/* neighbor history routines; don't fail on a send
   packet if the neighbor was there before. */
#define MAX_NEIGHBOR_HIST 1024
int neighbor_hist[MAX_NEIGHBOR_HIST];

static int find_in_neighbor_history(int naddr) {
  int i; 
  for(i=0; i<MAX_NEIGHBOR_HIST
        && neighbor_hist[i]!= 0
        && neighbor_hist[i]!= naddr ; 
      i++);
  if(neighbor_hist[i] == naddr) 
    return i;
  else
    return -1;
}
static void insert_neighbor_history(int naddr) {
  int i;
  for(i=0; i<MAX_NEIGHBOR_HIST
        && neighbor_hist[i]!= 0
        && neighbor_hist[i]!= naddr ; 
      i++);
  if(i < MAX_NEIGHBOR_HIST) {
    if(neighbor_hist[i] == 0) 
      neighbor_hist[i] = naddr;
  } else {
    printf("Non-fatal error: ran out of neighbor history space.\n");
  }
}

static void update_neighbor_history(void) {
  int i;
  for (i=0; i< MAX_NEIGHBORS && neighbors[i].address != 0 ; i++);
    insert_neighbor_history(neighbors[i].address);
}

/* used when the likely cause is system (possibly TA) error */
static void errorexit(const char* msg, int syserr) {
  fprintf(stderr, "%s\n",msg);
  if (syserr != 0) perror("\n fish ");
  exit(EXIT_FAILURE);
}

/* used when the likely cause is programmer (read: student) error */
void errorabort(const char* msg) {
  fprintf(stderr, "** %s **\n", msg);
  fprintf(stderr, "Generating core dump.  Use: \n"
          "  gdb [programname] core\n"
          "and use the commands: where, up, down, list, and print,\n"
          "if the error messages above (in **'s) haven't helped.\n");

  /* bail, generating core dump (if not limited, see getrlimit(2)) */
  abort();
}

static void fish_init(void) {
  (void)time(&last_send_time); 
  init = TRUE;
  srand((unsigned int)time(0));

  return;
}

extern const int* fish_getneighbors() {
  static int nlist[MAX_NEIGHBORS+1]; /* zero terminated */
  int i;

  if (!init) fish_init();
  memset(nlist, 0, sizeof(nlist));
  for (i=0; 
       neighbors[i].address != 0 && i < MAX_NEIGHBORS; 
       i++){
    nlist[i] = neighbors[i].address;
  }
  return nlist;
}

extern void fish_recvhook(receive_handler fn) {
  if (!init) fish_init();
  recvhandler = fn;
}

extern void fish_keybhook(keyboard_handler fn) {
  if (!init) fish_init();
  keybhandler = fn;
}

extern int fish_sendpacket(int naddr, void* pkt, int len) {
  int n;
  int i;
  int ok;
  struct sockaddr_in to;

  if (!init) fish_init();
  if(sockfishbits<0) {
    errorabort("fish: must join the network before sending a packet\n");
  }

  if (len > MAXPACKET) {
    errorabort("fish: packet too long\n");
  }

  packet_verification(pkt,len);

  n = -1;
  for (i = 0; i < MAX_NEIGHBORS; i++) {
    if (neighbors[i].address == naddr) n = i; 
  }
  if (n < 0) {
    /* only abort if this neighbor never existed. */
    if(find_in_neighbor_history(naddr) == -1) 
      return -1;
    else
      errorabort("fish: not a valid neighbor\n"); 
  }

  if ((unsigned int)((rand() >> 16 | rand()) % 1000) 
      < (unsigned int)neighbors[n].lossrate_x_1000) {
    fish_debug(6, "fish: Dropped frame from %d to %d\n",
               self_address, naddr );
    /* should be a silent failure. */
    return (0);
  }
  
  memset((void *)&to, 0, sizeof(to));
  to.sin_family      = AF_INET;
  to.sin_addr        = neighbors[n].node_ip_address;
  to.sin_port        = neighbors[n].node_udp_port;
  ok = sendto(sockfishbits, pkt, len, 0, (struct sockaddr*)&to, sizeof(to)); 
  if (ok < 0) {
    if(errno == ECONNREFUSED) {
      fish_debug(5, "fish: warning - neighbor %d seems to have failed\n", 
                 naddr);
    } else {
      fish_debug(0, "fish: while sending to %d at %s:%d\n", 
                 naddr, inet_ntoa(to.sin_addr), htons(to.sin_port));
      fish_debug(0, "fish: experienced error %d: %s\n", 
                 errno, strerror(errno));
      errorexit("fish: error, can't sendto in sendpacket\n", 1);
    }
  } else 
    fish_debug(FISH_DEBUG_ALL, "fish: Sent packet\n");
  
  usleep(10);
  return 0;
}

extern int fish_joinnetwork(const char* head_hostname, 
                            int head_port, 
                            int addr) {
  int ok;
  int address_length_buf ;

  struct hostent *hp;
  struct in_addr *firstAddress;
  char messagebuffer[1024];
  struct fish_message *pfm = (struct fish_message *)messagebuffer;
  struct joinrequest *jreq = &pfm->body.joinreq; 

  if (sockfishhead > 0
      || sockfishbits > 0) 
    errorabort("fish: error, already joined\n");

  if ( ! InRange(1, addr, MAX_NODE_ADDRESS)) {
    errorexit("fish: not a valid fishnet address", 0);
  }

  if (!init) fish_init();

  self_address = addr;

  // Open a UDP Port
  sockfishbits = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sockfishbits < 0) 
    errorexit("fish: error, can't get fishbits socket\n", 1);

  bzero((char *) &bitsaddr, sizeof(bitsaddr));
  bitsaddr.sin_family      = AF_INET;
  bitsaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  bitsaddr.sin_port        = htons(0);

  ok = bind(sockfishbits, (struct sockaddr*)&bitsaddr, sizeof(bitsaddr));
  if (ok < 0) 
    errorexit("fish: error, can't bind fishbits socket\n", 1);

  address_length_buf = sizeof(bitsaddr);
  ok = getsockname(sockfishbits,
                   (struct sockaddr *)&bitsaddr, 
                   &address_length_buf);
  if (ok < 0) 
    errorexit("fish: error, can't find port number of fishbits socket\n", 1);

  /// Code handling the network manager
  sockfishhead = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sockfishhead < 0) 
    errorexit("fish: error, can't get fishhead socket\n", 0);

  // Convert hostname to  ipaddress
  if ( (hp = gethostbyname(head_hostname)) == NULL) 
    errorexit(" Failure in mapping host name to IP",0);
  firstAddress = (struct in_addr *) hp->h_addr_list[0];
  if (firstAddress == NULL) 
    errorexit(" Failure in mapping host name to IP",0);
  bzero((char *) &headaddr, sizeof(headaddr));
  memcpy(&headaddr.sin_addr,firstAddress,sizeof(struct in_addr));
  headaddr.sin_family      = AF_INET;
  headaddr.sin_port        = htons(head_port);
  
  // Connect to the network manager
  ok = connect(sockfishhead,
               (struct sockaddr*) &headaddr,
               sizeof(headaddr));
  if (ok < 0) 
    errorexit("fish: error, can't get fishhead connection\n", 1);

  // Send a join request to the network manager
  pfm->type   = FISH_MSG_TYPE_JOINREQUEST;
  pfm->length = MIN_MESSAGE_SIZE + (sizeof(struct joinrequest));
  jreq->node_address = addr;
  jreq->node_udp_port = bitsaddr.sin_port;
  strncpy(jreq->username, getenv("USER"), 8);
  
  (void)write(sockfishhead, messagebuffer, (size_t)pfm->length);
 
  /* Now wait to hear the return message from the server
   * telling you your neighbor list... */
  (void)read_neighbor_list();
  printneighbors(neighbors);
  
  fish_debug(6, "fish: joined network %s:%i as %i\n", 
                 head_hostname, head_port, addr);
  return 0;
}

extern void fish_main(void) {
 
  if (!init) fish_init();

  if(sockfishbits<0) {
    errorabort("fish: must join the network before calling fish_main\n");
  }

  fish_debug(7, "fish: in main\n");
  
  while (1) {
    fd_set rd;
    int max = 0;
    struct timeval wait;
    int ok;

    FD_ZERO(&rd);
    FD_SET(sockfishhead, &rd);
    FD_SET(sockfishbits, &rd);
    FD_SET(STDIN_FILENO, &rd); 
    max = fish_constructmask(&rd, 
                              MAX3(sockfishhead, 
                                   sockfishbits, 
                                   STDIN_FILENO));
    ok = select(max+1, &rd, 0, 0, 
                fish_time_to_next_event(&wait) ? &wait : 0);
    if (ok < 0) {
      if(errno != EINTR) {
        errorexit("fish: error, select\n", 1);
      } else {
        continue;
      }
    }

    fish_executepending();

    if (FD_ISSET(STDIN_FILENO, &rd)) {
      char buf[1024];

      bzero((char *)&buf, sizeof(buf));

      fgets(buf,sizeof(buf),stdin);
      if(feof(stdin)) {
        fish_debug(FISH_DEBUG_NONE, "fish: bye!\n");
	      exit(EXIT_SUCCESS);
      }

      fish_debug(FISH_DEBUG_ALL, "fish: received keyboard input\n");
      if (keybhandler != NULL)
        (*keybhandler)(buf, strlen(buf));
    }
    
    if (FD_ISSET(sockfishbits, &rd)) {
      char buf[1024];
      int msglen;
      struct sockaddr_in from;
      int fromlen = sizeof(struct sockaddr_in);
      int fromid;

      bzero((char *)&buf, sizeof(buf));
      msglen = recvfrom(sockfishbits, buf, sizeof(buf), 
           0, (struct sockaddr*)&from, &fromlen);
      fromid = find_by_sock_address(from);
      fish_debug(FISH_DEBUG_ALL, "fish: received packet from %d\n", fromid);
      if (recvhandler != NULL)
        (*recvhandler)(fromid,buf, msglen);
    }
    
    if (FD_ISSET(sockfishhead, &rd)) {
      /* Update the neighbor list  */
      fish_debug(FISH_DEBUG_TOPOLOGY, "Neighbor update received.\n");
      (void)read_neighbor_list();
      printneighbors(neighbors);
    }

    /* handle anything else that was setup to use normal sockets */
    fish_processbindings(&rd);
  }
    
  return;
}


int find_by_sock_address(const struct sockaddr_in addr) {

  int i ;
  
  for(i=0; i< MAX_NEIGHBORS; i++) {
    if ((neighbors[i].node_ip_address.s_addr == addr.sin_addr.s_addr)
        && (neighbors[i].node_udp_port == addr.sin_port)) {
      return neighbors[i].address;
    }
  }
  return -1;

}

/* helper function for packet_verification (below) that 
   verifies that the length of a text containing packet
   (either assignment 1 style or ECHO/ECHO_RESPONSE is no
   larger than necessary */  
static boolean checkStringLength(const char *string, int string_length) {
  /* strnlen is a gnu-specific function, easily re-implemented */
  if(string_length != 0 && strnlen(string, string_length) < string_length-1) {
    fprintf(stderr, "The frame containing the following text:\n"
            "%*s\n"
            "was sent with a length of %d bytes, "
            "but only requires %d bytes\n", 
            string_length, string, 
            string_length, strnlen(string,string_length));
    return(FALSE);
  } else 
    return(TRUE);
}

/* abort()s on failure */
void packet_verification(const char *inp, const int packet_length) {
  int i;
  int textframe;
  int toCheckLen;
  const struct packet *pkt = (const struct packet *)inp;

  if(inp == NULL) 
    errorabort("Null packet");

  if (packet_length > MTU) 
    errorabort("Size of packet > MTU");
  if (packet_length < 0) 
    errorabort("Packet has negative size");
  
  // Check if the packet is text only...
  toCheckLen = MIN(16, packet_length);
  for(i=0, textframe = 1; i< toCheckLen; i++) {
    textframe &= (isprint(inp[i]) || isspace(inp[i])) ;
  }
  
  if (textframe) {
    /* Verify that the frame length (the function parameter)
       is at most as long as the string plus a safe
       terminating zero.  This is not an egregious error,
       but verifying it here is probably the only way to
       make sure students pass a length parameter that is
       not too long. */
    if(!checkStringLength(inp, packet_length)) {
      errorabort("Text frame was too long");
    }
    return;
  }
  
  // Packet is not just text
  if ( ! InRange(1, pkt->src, MAX_NODE_ADDRESS)) {
    PRINT_AND_BAIL("In send_packet: Invalid Source");
  }
  
  if ( ! InRange(1, pkt->dst, MAX_NODE_ADDRESS)) {
    PRINT_AND_BAIL("In send_packet: Invalid Destination");
  }
  if ( ! InRange(1, pkt->ttl, MAX_TTL)) {
    PRINT_AND_BAIL("In send_packet: Invalid TTL");
  }

  if ( ! InRange(1, pkt->protocol, FISH_PROTOCOL_MAX)) {
    PRINT_AND_BAIL("In send_packet:Invalid Protcol");
  }
  
  if (pkt->dst == self_address) {
    PRINT_AND_BAIL("In send_packet: Destination of the packet is same as the source");
  }

  // Protocols specific checkings...
  if (pkt->protocol == FISH_PROTOCOL_ROUTING){
    if (pkt->ttl != 1)  {
      PRINT_AND_BAIL("In send_packet: TTL must be 1 one for a Routing Packet");
    }
    if (pkt->src != self_address) {
      PRINT_AND_BAIL("In send_packet: Source for a Routing Packet must be same as self");
    }
    /* Check if the destination is one of the
       neighbors.  There is no broadcast address */
    for (i = 0; i < MAX_NEIGHBORS && neighbors[i].address != pkt->dst; i++);
    if (i == MAX_NEIGHBORS) {
      PRINT_AND_BAIL("In send_packet: Destination of a Routing Packet must be a neighbor.");
    }
  }

  
  if (pkt->protocol == FISH_PROTOCOL_TRANSPORT){
    const struct transport_packet *tpkt = 
      (const struct transport_packet *)inp;

    if ( ! InRange(0,tpkt->src_port,MAX_PORT_NUMBER)) {
      PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid src port");
    }
    if ( ! InRange(0,tpkt->dst_port,MAX_PORT_NUMBER)) {
      PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid dst port");
    }
    
    if(!((tpkt->fin==0) || (tpkt->fin ==1))) {
      PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid fin (must be zero or one)");
    }
     
    if(!((tpkt->syn==0) || (tpkt->syn ==1))) {
      PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid syn (must be zero or one)");
    }
    

    if (tpkt->seq_no == 0) {
      if (!InRange(1,tpkt->ack_no,MAX_SEQ_NO)) {
	PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid ack no");
      }
    }
    else if (tpkt->ack_no == 0) {
      if (!InRange(1,tpkt->seq_no,MAX_SEQ_NO)) {
	PRINT_AND_BAIL("In send_packet:Transport protocol packet has invalid seq no");
      }
    }
    else {
      // this is ok -nspring; printpacket(pkt,packet_length);
      // both should be set in a data packet. errorabort("In send_packet:Transport protocol packet,both seq no and ack not found to be zero");
    }
  }
    
  if (pkt->protocol == FISH_PROTOCOL_ECHO 
      || pkt->protocol == FISH_PROTOCOL_ECHO_RESPONSE) {
    if(!checkStringLength(pkt->data, packet_length-PACKET_HEADER_SIZE)) {
      PRINT_AND_BAIL("In send_packet: ECHO/ECHO_RESPONSE frame was too long");
    }
  }
}


void printpacket(const struct packet *pkt, int length)  {
  static const char *strpacket[] = {
    "Error",
    "Echo",
    "", 
    "Echo response",
    "Routing",
    "Transport"
  };
  if(pkt) {
    if(pkt->protocol != FISH_PROTOCOL_TRANSPORT) {
      fprintf(stderr,"Packet: [src,dst,ttl,protocol,data] [%d,%d,%d,%s,%s]\n", 
              pkt->src, pkt->dst, pkt->ttl,
              strpacket[ pkt->protocol], pkt->data);
    } else {
      struct transport_packet *tpkt = (struct transport_packet *)pkt;
      int datalength = 
        MAX(length - TRANSPORT_HEADER_SIZE - PACKET_HEADER_SIZE,
                       0);
      fprintf(stderr,"%d:%d -> %d:%d %d ack %d ttl:%d %s %s len:%d data:%.*s\n", 
              tpkt->src, tpkt->src_port, 
              tpkt->dst, tpkt->dst_port,
              tpkt->seq_no, tpkt->ack_no,
              tpkt->ttl,
              tpkt->fin ? "fin" : "", 
	      tpkt->syn ? "syn" : "", 
              datalength,
              datalength,
              /* convoluted - print if nonzero and not totally binary */
              datalength ? (isprint(tpkt->data[0]) ? tpkt->data : "binary") : "");
    }
  } else {
    fprintf(stderr,"Packet: null pointer\n");
  }
}

void read_neighbor_list() {
#define MSG_BUF_SZ 1024
  char messagebuffer[MSG_BUF_SZ];
  int readlen;

  struct fish_message *pfm = (struct fish_message *)messagebuffer;

  bzero((char *) &messagebuffer, sizeof(messagebuffer));
  readlen = read(sockfishhead, messagebuffer, MSG_BUF_SZ);
  if (readlen <= 0) {
    errorexit("Fishead died\n",0);
  }

  if(pfm->length > readlen) {
    errorexit("read() failed, message not completely read",0);
  }
  
  if (pfm->type != FISH_MSG_TYPE_NEIGHBORLIST){
    printmessage(pfm);
    errorabort("Above Message received from fish-head \n");
  }
  else {
    int i;
    // printmessage(pfm);
    for(i=0; i< MAX_NEIGHBORS; i++) {
     neighbors[i].address = pfm->body.neighbors[i].address ;
     neighbors[i].node_ip_address =  pfm->body.neighbors[i].node_ip_address ;
     neighbors[i].node_udp_port = pfm->body.neighbors[i].node_udp_port ;
     neighbors[i].lossrate_x_1000 = pfm->body.neighbors[i].lossrate_x_1000;
     assert(neighbors[i].lossrate_x_1000 <= 1000);
    }
    update_neighbor_history();
  }
}

static struct select_binding {
  struct select_binding *next;
  int sd;
  void (*handler)(int);
} *binding_list;

void fish_internal_readhook(int sd, void (*rdfn)(int)) {
  struct select_binding *psb = 
    (struct select_binding *)malloc(sizeof(struct select_binding));
  psb->sd = sd;
  psb->handler = rdfn;
  psb->next = binding_list;
  binding_list = psb;
}

void fish_internal_removereadhook(int sd) {
  struct select_binding *psb, *qsb;
  for(psb = binding_list, qsb = NULL; 
      psb != NULL && psb->sd != sd;
      qsb=psb, psb=psb->next);

  if(psb != NULL) { /* found it */
    if(qsb != NULL) { /* in the middle */
      qsb->next = psb->next;
    } else { /* at the beginning */
      binding_list = psb->next;
    }
    free(psb);
  }
}

static int fish_constructmask(fd_set *prd, int max) {
  struct select_binding *psb;
  for(psb = binding_list; 
      psb != NULL; 
      psb = psb->next) {
    FD_SET(psb->sd, prd); 
    max = MAX(psb->sd, max);
  }
  return(max);
}

static void fish_processbindings(fd_set *prd) {
  struct select_binding *psb, *next;
  for(psb = binding_list; psb != NULL; psb = next) {
    next = psb->next; /* grab this now, in case we remove
                         ourselves from the list */
    if(psb->sd > 0 && psb->sd < 4096) {
      if(FD_ISSET(psb->sd, prd)) {
        psb->handler(psb->sd);
        return; 
      }
    } else {
      printf("something got corrupted\n");
    }
  }
}


